"""
A trainer class to handle training and testing of models.
"""

import sys

import torch
from phonlp.models.common import utils as util
from phonlp.models.common.chuliu_edmonds import chuliu_edmonds_one_root
from phonlp.models.common.crf import viterbi_decode
from phonlp.models.common.trainer import Trainer as BaseTrainer
from phonlp.models.jointmodel.model import JointModel
from phonlp.models.ner.vocab import MultiVocab


def unpack_batch(batch, use_cuda, type):
    """ Unpack a batch from the data loader. """
    if type == "dep":
        if use_cuda:
            inputs = [b.cuda() if b is not None else None for b in batch[:5]]
        else:
            inputs = batch[:5]
        number_of_words = batch[5]
        orig_idx = batch[6]
        sentlens = batch[7]
        return inputs, number_of_words, orig_idx, sentlens
    elif type == "pos":
        if use_cuda:
            inputs = [b.cuda() if b is not None else None for b in batch[:3]]
        else:
            inputs = batch[:3]
        orig_idx = batch[3]
        sentlens = batch[4]
    elif type == "ner":
        if use_cuda:
            inputs = [b.cuda() if b is not None else None for b in batch[:4]]
        else:
            inputs = batch[:4]
        orig_idx = batch[4]
        sentlens = batch[5]
    return inputs, orig_idx, sentlens


class JointTrainer(BaseTrainer):
    """ A trainer for training models. """

    def __init__(self, args=None, vocab=None, model_file=None, config_phobert=None, use_cuda=False):
        self.use_cuda = use_cuda
        self.config_phobert = config_phobert
        if model_file is not None:
            self.load(model_file)
        else:
            self.args = args
            self.vocab = vocab
            self.model = JointModel(args, vocab, self.config_phobert)

        if self.use_cuda:
            self.model.cuda()
        else:
            self.model.cpu()

    def update(self, batch_dep=None, batch_pos=None, batch_ner=None, lambda_pos=1.0, lambda_ner=1.0, lambda_dep=1.0):
        dep_inputs, dep_number_of_words, dep_orig_idx, dep_sentlens = unpack_batch(
            batch_dep, self.use_cuda, type="dep"
        )
        dep_tokens_phobert, dep_first_subword, dep_words_mask, dep_head, dep_deprel = dep_inputs
        pos_inputs, pos_orig_idx, pos_sentlens = unpack_batch(batch_pos, self.use_cuda, type="pos")
        pos_tokens_phobert, pos_first_subword, pos_upos = pos_inputs
        ner_inputs, ner_orig_idx, ner_sentlens = unpack_batch(batch_ner, self.use_cuda, type="ner")
        ner_tokens_phobert, ner_first_subword, ner_words_mask, ner_tags = ner_inputs

        self.model.train()
        loss_pos, loss_ner, loss_dep, loss, _ = self.model.forward(
            dep_tokens_phobert,
            dep_first_subword,
            dep_words_mask,
            dep_number_of_words,
            dep_head,
            dep_deprel,
            dep_sentlens,
            pos_tokens_phobert,
            pos_first_subword,
            pos_upos,
            pos_sentlens,
            ner_tokens_phobert,
            ner_first_subword,
            ner_words_mask,
            ner_sentlens,
            ner_tags,
            lambda_pos,
            lambda_ner,
            lambda_dep,
        )
        loss_val = loss.data.item()
        loss_val_pos = loss_pos.data.item()
        loss_val_ner = loss_ner.data.item()
        loss_val_dep = loss_dep.data.item()
        loss.backward()
        return loss_val, loss_val_pos, loss_val_ner, loss_val_dep

    def predict_dep(self, batch, unsort=True):
        dep_inputs, dep_number_of_words, dep_orig_idx, dep_sentlens = unpack_batch(batch, self.use_cuda, type="dep")
        dep_tokens_phobert, dep_first_subword, dep_words_mask, dep_head, dep_deprel = dep_inputs
        self.model.eval()
        batch_size = dep_tokens_phobert.size(0)
        loss_dep, preds = self.model.dep_forward(
            dep_tokens_phobert,
            dep_first_subword,
            dep_words_mask,
            dep_number_of_words,
            dep_sentlens,
            dep_head,
            dep_deprel,
            eval=True,
        )
        # dependency
        head_seqs = [
            chuliu_edmonds_one_root(adj[:l, :l])[1:] for adj, l in zip(preds[0], dep_sentlens)
        ]  # remove attachment for the root
        deprel_seqs = [
            self.vocab["deprel"].unmap([preds[1][i][j + 1][h] for j, h in enumerate(hs)])
            for i, hs in enumerate(head_seqs)
        ]
        pred_tokens = [
            [[str(head_seqs[i][j]), deprel_seqs[i][j]] for j in range(dep_sentlens[i] - 1)] for i in range(batch_size)
        ]
        if unsort:
            pred_tokens = util.unsort(pred_tokens, dep_orig_idx)
        return pred_tokens

    def predict_pos(self, batch, unsort=True):
        pos_inputs, pos_orig_idx, pos_sentlens = unpack_batch(batch, self.use_cuda, type="pos")
        pos_tokens_phobert, pos_first_subword, pos_upos = pos_inputs
        self.model.eval()
        batch_size = pos_tokens_phobert.size(0)
        _, preds = self.model.pos_forward(pos_tokens_phobert, pos_first_subword, pos_sentlens, False, pos_upos)
        upos_seqs = [self.vocab["upos"].unmap(sent) for sent in preds[0].tolist()]

        pred_tokens = [
            [[upos_seqs[i][j]] for j in range(pos_sentlens[i])] for i in range(batch_size)
        ]  # , xpos_seqs[i][j], feats_seqs[i][j]
        if unsort:
            pred_tokens = util.unsort(pred_tokens, pos_orig_idx)
        return pred_tokens

    def predict_ner(self, batch, unsort=True):
        ner_inputs, ner_orig_idx, ner_sentlens = unpack_batch(batch, self.use_cuda, type="ner")
        ner_tokens_phobert, ner_first_subword, ner_word_mask, ner_tags = ner_inputs

        self.model.eval()
        loss, logits = self.model.ner_forward(
            ner_tokens_phobert, ner_first_subword, ner_word_mask, ner_sentlens, ner_tags
        )

        # decode
        trans = self.model.crit_ner._transitions.data.cpu().numpy()
        scores = logits.data.cpu().numpy()
        bs = logits.size(0)
        tag_seqs = []
        for i in range(bs):
            tags, _ = viterbi_decode(scores[i, : ner_sentlens[i]], trans)
            tags = self.vocab["ner_tag"].unmap(tags)
            tag_seqs += [tags]
        if unsort:
            tag_seqs = util.unsort(tag_seqs, ner_orig_idx)
        return tag_seqs

    def save(self, filename):
        model_state = self.model.state_dict()
        # skip saving modules like pretrained embeddings, because they are large and will be saved in a separate file
        params = {"model": model_state, "vocab": self.vocab.state_dict(), "config": self.args}
        try:
            torch.save(params, filename)
            print("Model saved to {}".format(filename))
        except BaseException:
            print("Saving failed... continuing anyway.")

    def load(self, filename):
        """
        Load a model from file
        """
        try:
            checkpoint = torch.load(filename, lambda storage, loc: storage)
        except BaseException:
            print("Cannot load model from {}".format(filename))
            sys.exit(1)
        self.args = checkpoint["config"]
        self.vocab = MultiVocab.load_state_dict(checkpoint["vocab"])
        # load model
        self.model = JointModel(self.args, self.vocab, config=self.config_phobert)
        self.model.load_state_dict(checkpoint["model"], strict=False)



"""
Constants for seq2seq models.
"""

PAD = "<PAD>"
PAD_ID = 0
UNK = "<UNK>"
UNK_ID = 1
SOS = "<SOS>"
SOS_ID = 2
EOS = "<EOS>"
EOS_ID = 3

VOCAB_PREFIX = [PAD, UNK, SOS, EOS]

EMB_INIT_RANGE = 1.0
INFINITY_NUMBER = 1e12



"""
An NER scorer that calculates F1 score given gold and predicted tags.
"""
import logging
from collections import Counter

from phonlp.models.ner.utils import decode_from_bioes


logger = logging.getLogger("PhoNLPToolkit")


def score_by_entity(pred_tag_sequences, gold_tag_sequences, verbose=True):
    """Score predicted tags at the entity level.

    Args:
        pred_tags_sequences: a list of list of predicted tags for each word
        gold_tags_sequences: a list of list of gold tags for each word
        verbose: print log with results

    Returns:
        Precision, recall and F1 scores.
    """
    assert len(gold_tag_sequences) == len(
        pred_tag_sequences
    ), "Number of predicted tag sequences does not match gold sequences."

    def decode_all(tag_sequences):
        # decode from all sequences, each sequence with a unique id
        ents = []
        for sent_id, tags in enumerate(tag_sequences):
            for ent in decode_from_bioes(tags):
                ent["sent_id"] = sent_id
                ents += [ent]
        return ents

    gold_ents = decode_all(gold_tag_sequences)
    pred_ents = decode_all(pred_tag_sequences)

    # scoring
    correct_by_type = Counter()
    guessed_by_type = Counter()
    gold_by_type = Counter()

    for p in pred_ents:
        guessed_by_type[p["type"]] += 1
        if p in gold_ents:
            correct_by_type[p["type"]] += 1
    for g in gold_ents:
        gold_by_type[g["type"]] += 1

    prec_micro = 0.0
    if sum(guessed_by_type.values()) > 0:
        prec_micro = sum(correct_by_type.values()) * 1.0 / sum(guessed_by_type.values())
    rec_micro = 0.0
    if sum(gold_by_type.values()) > 0:
        rec_micro = sum(correct_by_type.values()) * 1.0 / sum(gold_by_type.values())
    f_micro = 0.0
    if prec_micro + rec_micro > 0:
        f_micro = 2.0 * prec_micro * rec_micro / (prec_micro + rec_micro)

    if verbose:
        logger.info("Prec.\tRec.\tF1")
        logger.info("{:.2f}\t{:.2f}\t{:.2f}".format(prec_micro * 100, rec_micro * 100, f_micro * 100))
    return prec_micro, rec_micro, f_micro


def score_by_token(pred_tag_sequences, gold_tag_sequences, verbose=True):
    """Score predicted tags at the token level.

    Args:
        pred_tags_sequences: a list of list of predicted tags for each word
        gold_tags_sequences: a list of list of gold tags for each word
        verbose: print log with results

    Returns:
        Precision, recall and F1 scores.
    """
    assert len(gold_tag_sequences) == len(
        pred_tag_sequences
    ), "Number of predicted tag sequences does not match gold sequences."

    correct_by_tag = Counter()
    guessed_by_tag = Counter()
    gold_by_tag = Counter()

    for gold_tags, pred_tags in zip(gold_tag_sequences, pred_tag_sequences):
        assert len(gold_tags) == len(pred_tags), "Number of predicted tags does not match gold."
        for g, p in zip(gold_tags, pred_tags):
            if g == "O" and p == "O":
                continue
            elif g == "O" and p != "O":
                guessed_by_tag[p] += 1
            elif g != "O" and p == "O":
                gold_by_tag[g] += 1
            else:
                guessed_by_tag[p] += 1
                gold_by_tag[p] += 1
                if g == p:
                    correct_by_tag[p] += 1

    prec_micro = 0.0
    if sum(guessed_by_tag.values()) > 0:
        prec_micro = sum(correct_by_tag.values()) * 1.0 / sum(guessed_by_tag.values())
    rec_micro = 0.0
    if sum(gold_by_tag.values()) > 0:
        rec_micro = sum(correct_by_tag.values()) * 1.0 / sum(gold_by_tag.values())
    f_micro = 0.0
    if prec_micro + rec_micro > 0:
        f_micro = 2.0 * prec_micro * rec_micro / (prec_micro + rec_micro)

    if verbose:
        logger.info("Prec.\tRec.\tF1")
        logger.info("{:.2f}\t{:.2f}\t{:.2f}".format(prec_micro * 100, rec_micro * 100, f_micro * 100))
    return prec_micro, rec_micro, f_micro


def test():
    pred_sequences = [
        ["O", "S-LOC", "O", "O", "B-PER", "E-PER"],
        ["O", "S-MISC", "O", "E-ORG", "O", "B-PER", "I-PER", "E-PER"],
    ]
    gold_sequences = [
        ["O", "B-LOC", "E-LOC", "O", "B-PER", "E-PER"],
        ["O", "S-MISC", "B-ORG", "E-ORG", "O", "B-PER", "E-PER", "S-LOC"],
    ]
    print(score_by_token(pred_sequences, gold_sequences))
    print(score_by_entity(pred_sequences, gold_sequences))


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
import gdown
import torch
from phonlp.annotate_model import JointModel
from phonlp.models.common import utils as util
from phonlp.models.ner.vocab import MultiVocab
from transformers import AutoConfig, AutoTokenizer


def download(save_dir, url="https://drive.google.com/uc?id=1wfVGVsKkg7sH8R_3ECRoQ46DVOAmsNkQ"):
    util.ensure_dir(save_dir)
    if save_dir[len(save_dir) - 1] == "/":
        model_file = save_dir + "phonlp.pt"
    else:
        model_file = save_dir + "/phonlp.pt"
    gdown.download(url, model_file)


def load(save_dir="./"):
    if save_dir[len(save_dir) - 1] == "/":
        model_file = save_dir + "phonlp.pt"
    else:
        model_file = save_dir + "/phonlp.pt"
    print("Loading model from: {}".format(model_file))
    checkpoint = torch.load(model_file, lambda storage, loc: storage)
    args = checkpoint["config"]
    vocab = MultiVocab.load_state_dict(checkpoint["vocab"])
    # load model
    tokenizer = AutoTokenizer.from_pretrained(args["pretrained_lm"], use_fast=False)
    config_phobert = AutoConfig.from_pretrained(args["pretrained_lm"], output_hidden_states=True)
    model = JointModel(args, vocab, config_phobert, tokenizer)
    model.load_state_dict(checkpoint["model"], strict=False)
    if torch.cuda.is_available() is False:
        model.to(torch.device("cpu"))
    else:
        model.to(torch.device("cuda"))
    model.eval()
    return model


if __name__ == "__main__":
    download("./")
    model = load("./")
    text = "Tôi tên là Thế_Linh ."
    output = model.annotate(text=text)
    model.print_out(output)



"""
Utils and wrappers for scoring parsers.
"""
import logging

from phonlp.models.common.utils import ud_scores


logger = logging.getLogger("PhoNLPToolkit")


def score(system_conllu_file, gold_conllu_file, verbose=True):
    """ Wrapper for UD parser scorer. """
    evaluation = ud_scores(gold_conllu_file, system_conllu_file)
    el = evaluation["LAS"]
    el2 = evaluation["UAS"]
    p = el.precision
    r = el.recall
    f = el.f1
    if verbose:
        scores = [evaluation[k].f1 * 100 for k in ["UAS", "LAS"]]
        logger.info("UAS\tLAS")
        logger.info("{:.2f}\t{:.2f}".format(*scores))
    return p, r, f, el2.f1



# -*- coding: utf-8 -*-
"""
Training, evaluation and annotate for joint model.
"""

import argparse
import random
import sys

import numpy as np
import torch
from phonlp.annotate_model import JointModel
from phonlp.models.common import utils as util
from phonlp.models.common.doc import DEPREL, HEAD, Document
from phonlp.models.depparse import scorer as score_dep
from phonlp.models.jointmodel.data import BuildVocab, DataLoaderDep, DataLoaderNER, DataLoaderPOS
from phonlp.models.jointmodel.trainer import JointTrainer
from phonlp.models.ner import scorer as score_ner
from phonlp.models.ner.vocab import MultiVocab
from phonlp.models.pos import scorer as score_pos
from phonlp.utils.conll import CoNLL
from tqdm import tqdm
from transformers import AdamW, AutoConfig, AutoTokenizer, get_constant_schedule, get_linear_schedule_with_warmup


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--train_file_dep",
        type=str,
        default="/home/ubuntu/linhnt140/data/VnDTv1.1_predictedPOS/VnDTv1.1-train.conll",
        help="Input file for data loader.",
    )

    parser.add_argument(
        "--eval_file_dep",
        type=str,
        default="/home/ubuntu/linhnt140/data/VnDTv1.1_predictedPOS/VnDTv1.1-dev.conll",
        help="Input file for data loader.",
    )
    parser.add_argument("--output_file_dep", type=str, default="./jointmodel/dep.out", help="Output CoNLL-U file.")

    # POS
    parser.add_argument(
        "--train_file_pos",
        type=str,
        default="/home/ubuntu/linhnt140/data/POS_data/POS_data/VLSP2013_POS_train.txt",
        help="Input file for data loader.",
    )
    parser.add_argument(
        "--eval_file_pos",
        type=str,
        default="/home/ubuntu/linhnt140/data/POS_data/POS_data/VLSP2013_POS_dev.txt",
        help="Input file for data loader.",
    )

    # NER
    parser.add_argument(
        "--train_file_ner",
        type=str,
        default="/home/ubuntu/linhnt140/data/NER_data/train.txt",
        help="Input file for data loader.",
    )
    parser.add_argument(
        "--eval_file_ner",
        type=str,
        default="/home/ubuntu/linhnt140/data/NER_data/dev.txt",
        help="Input file for data loader.",
    )
    # Anotate corpus
    parser.add_argument(
        "--input_file",
        type=str,
        default="/home/ubuntu/linhnt140/data/annotate/input.txt",
        help="Input file for annotate corpus.",
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default="/home/ubuntu/linhnt140/data/annotate/output.txt",
        help="Output file for annotate corpus.",
    )

    parser.add_argument("--mode", default="train", choices=["train", "eval", "annotate"])
    parser.add_argument("--deep_biaff_hidden_dim", type=int, default=400)
    parser.add_argument("--tag_emb_dim", type=int, default=100)
    parser.add_argument("--word_dropout", type=float, default=0.33)
    parser.add_argument("--dropout", type=float, default=0.5)
    parser.add_argument("--linearization", type=bool, default=True, help="Turn off linearization term.")
    parser.add_argument("--distance", type=bool, default=True, help="Turn off linearization term.")

    parser.add_argument("--lr", type=float, default=1e-5, help="Learning rate")
    parser.add_argument("--num_epoch", type=int, default=40)

    parser.add_argument("--lambda_pos", type=float, default=0.4, help="weight for pos loss.")
    parser.add_argument("--lambda_ner", type=float, default=0.2, help="weight for ner loss.")
    parser.add_argument("--lambda_dep", type=float, default=0.4, help="weight for dep loss.")

    parser.add_argument("--eval_interval", type=int, default=200)
    parser.add_argument("--batch_size", type=int, default=32)
    parser.add_argument("--accumulation_steps", type=int, default=1)
    parser.add_argument("--save_dir", type=str, default="saved_models/jointmodel", help="Root dir for saving models.")

    parser.add_argument("--seed", type=int, default=1234)
    parser.add_argument("--cuda", type=bool, default=torch.cuda.is_available())
    parser.add_argument("--cpu", action="store_true", help="Ignore CUDA.")
    # bert
    parser.add_argument("--pretrained_lm", type=str, default="vinai/phobert-base")
    parser.add_argument("--max_sequence_length", type=int, default=256)
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    torch.manual_seed(args.seed)
    np.random.seed(args.seed)
    random.seed(args.seed)
    if args.cpu:
        args.cuda = False
    elif args.cuda:
        torch.cuda.manual_seed(args.seed)

    args = vars(args)

    if args["mode"] == "train":
        train(args)
    elif args["mode"] == "eval":
        evaluate(args)
    else:
        annotate(
            input_file=args["input_file"], output_file=args["output_file"], args=args, batch_size=args["batch_size"]
        )


def train(args):
    util.ensure_dir(args["save_dir"])
    model_file = args["save_dir"] + "/" + "phonlp.pt"

    tokenizer = AutoTokenizer.from_pretrained(args["pretrained_lm"], use_fast=False)
    config_phobert = AutoConfig.from_pretrained(args["pretrained_lm"], output_hidden_states=True)

    print("Loading data with batch size {}...".format(args["batch_size"]))
    train_doc_dep = Document(CoNLL.conll2dict(input_file=args["train_file_dep"]))
    vocab = BuildVocab(args, args["train_file_pos"], train_doc_dep, args["train_file_ner"]).vocab

    train_batch_pos = DataLoaderPOS(
        args["train_file_pos"],
        args["batch_size"],
        args,
        vocab=vocab,
        evaluation=False,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    train_batch_dep = DataLoaderDep(
        train_doc_dep,
        args["batch_size"],
        args,
        vocab=vocab,
        evaluation=False,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    train_batch_ner = DataLoaderNER(
        args["train_file_ner"],
        args["batch_size"],
        args,
        vocab=vocab,
        evaluation=False,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )

    dev_doc_dep = Document(CoNLL.conll2dict(input_file=args["eval_file_dep"]))

    dev_batch_pos = DataLoaderPOS(
        args["eval_file_pos"],
        args["batch_size"],
        args,
        vocab=vocab,
        sort_during_eval=True,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    dev_batch_dep = DataLoaderDep(
        dev_doc_dep,
        args["batch_size"],
        args,
        vocab=vocab,
        sort_during_eval=True,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    dev_batch_ner = DataLoaderNER(
        args["eval_file_ner"],
        args["batch_size"],
        args,
        vocab=vocab,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )

    # pred and gold path
    system_pred_file = args["output_file_dep"]
    gold_file = args["eval_file_dep"]

    # ##POS

    dev_gold_tags = dev_batch_ner.tags

    # skip training if the language does not have training or dev data
    if len(train_batch_pos) == 0 or len(dev_batch_pos) == 0:
        print("Skip training because no data available...")
        sys.exit(0)

    print("Training jointmodel...")
    trainer = JointTrainer(args, vocab, None, config_phobert, args["cuda"])
    # ###
    tsfm = trainer.model.phobert
    for child in tsfm.children():
        for param in child.parameters():
            if not param.requires_grad:
                print("whoopsies")
            param.requires_grad = True

    global_step = 0
    las_score_history = 0
    uas_score_history = 0
    upos_score_history = 0
    f1_score_history = 0
    ####

    # start training
    train_loss = 0
    train_loss_pos = 0
    train_loss_dep = 0
    train_loss_ner = 0

    # Creating optimizer and lr schedulers
    param_optimizer = list(trainer.model.named_parameters())
    no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {"params": [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], "weight_decay": 0.01},
        {"params": [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], "weight_decay": 0.0},
    ]
    num_train_optimization_steps = int(args["num_epoch"] * len(train_batch_pos) / args["accumulation_steps"])
    optimizer = AdamW(
        optimizer_grouped_parameters, lr=args["lr"], correct_bias=False
    )  # To reproduce BertAdam specific behavior set correct_bias=False
    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=5, num_training_steps=num_train_optimization_steps
    )
    get_constant_schedule(optimizer)
    for epoch in range(args["num_epoch"]):
        ####
        optimizer.zero_grad()
        print(" EPOCH  : ", epoch)
        step = 0
        lambda_pos = args["lambda_pos"]
        lambda_ner = args["lambda_ner"]
        lambda_dep = args["lambda_dep"]

        epoch_size = max([len(train_batch_pos), len(train_batch_dep), len(train_batch_ner)])
        for i in tqdm(range(epoch_size)):
            step += 1
            global_step += 1
            batch_pos = train_batch_pos[i]
            batch_dep = train_batch_dep[i]
            batch_ner = train_batch_ner[i]
            ###
            loss, loss_pos, loss_ner, loss_dep = trainer.update(
                batch_dep, batch_pos, batch_ner, lambda_pos=lambda_pos, lambda_dep=lambda_dep, lambda_ner=lambda_ner
            )  # update step
            train_loss += loss
            train_loss_pos += loss_pos
            train_loss_dep += loss_dep
            train_loss_ner += loss_ner
            ###

            if i % args["accumulation_steps"] == 0:
                optimizer.step()
                optimizer.zero_grad()
                scheduler.step()

            if epoch_size == len(train_batch_pos):
                if step % len(train_batch_dep) == 0:
                    train_batch_dep.reshuffle()
                if step % len(train_batch_ner) == 0:
                    train_batch_ner.reshuffle()
            elif epoch_size == len(train_batch_ner):
                if step % len(train_batch_dep) == 0:
                    train_batch_dep.reshuffle()
                if step % len(train_batch_pos) == 0:
                    train_batch_pos.reshuffle()
            elif epoch_size == len(train_batch_dep):
                if step % len(train_batch_pos) == 0:
                    train_batch_dep.reshuffle()
                if step % len(train_batch_ner) == 0:
                    train_batch_ner.reshuffle()
            if step % args["eval_interval"] == 0:
                print("Evaluating on dev set...")
                dev_preds_dep = []
                dev_preds_upos = []
                dev_preds_ner = []
                for batch in dev_batch_dep:
                    preds_dep = trainer.predict_dep(batch)
                    dev_preds_dep += preds_dep
                ###
                dev_preds_dep = util.unsort(dev_preds_dep, dev_batch_dep.data_orig_idx_dep)
                dev_batch_dep.doc_dep.set([HEAD, DEPREL], [y for x in dev_preds_dep for y in x])
                CoNLL.dict2conll(dev_batch_dep.doc_dep.to_dict(), system_pred_file)
                _, _, las_dev, uas_dev = score_dep.score(system_pred_file, gold_file)

                for batch in dev_batch_pos:
                    preds_pos = trainer.predict_pos(batch)
                    dev_preds_upos += preds_pos
                dev_preds_upos = util.unsort(dev_preds_upos, dev_batch_pos.data_orig_idx_pos)
                accuracy_pos_dev = score_pos.score_acc(dev_preds_upos, dev_batch_pos.upos)

                for batch in dev_batch_ner:
                    preds_ner = trainer.predict_ner(batch)
                    dev_preds_ner += preds_ner
                p, r, f1 = score_ner.score_by_entity(dev_preds_ner, dev_gold_tags)
                for i in range(len(dev_batch_ner)):
                    assert len(dev_preds_ner[i]) == len(dev_gold_tags[i])

                print(
                    "step {}: dev_las_score = {:.4f}, dev_uas_score = {:.4f}, dev_pos = {:.4f}, dev_ner_p = {:.4f}, dev_ner_r = {:.4f}, dev_ner_f1 = {:.4f}".format(
                        global_step, las_dev, uas_dev, accuracy_pos_dev, p, r, f1
                    )
                )

                # save best model
                if las_dev + accuracy_pos_dev + f1 >= (las_score_history + upos_score_history + f1_score_history):
                    las_score_history = las_dev
                    upos_score_history = accuracy_pos_dev
                    uas_score_history = uas_dev
                    f1_score_history = f1
                    trainer.save(model_file)
                    print("new best model saved.")
                print("")

        print("Evaluating on dev set...")
        dev_preds_dep = []
        dev_preds_upos = []
        dev_preds_ner = []
        for batch in dev_batch_dep:
            preds_dep = trainer.predict_dep(batch)
            dev_preds_dep += preds_dep

        dev_preds_dep = util.unsort(dev_preds_dep, dev_batch_dep.data_orig_idx_dep)
        dev_batch_dep.doc_dep.set([HEAD, DEPREL], [y for x in dev_preds_dep for y in x])
        CoNLL.dict2conll(dev_batch_dep.doc_dep.to_dict(), system_pred_file)
        _, _, las_dev, uas_dev = score_dep.score(system_pred_file, gold_file)

        for batch in dev_batch_pos:
            preds_pos = trainer.predict_pos(batch)
            dev_preds_upos += preds_pos
        dev_preds_upos = util.unsort(dev_preds_upos, dev_batch_pos.data_orig_idx_pos)
        accuracy_pos_dev = score_pos.score_acc(dev_preds_upos, dev_batch_pos.upos)

        for batch in dev_batch_ner:
            preds_ner = trainer.predict_ner(batch)
            dev_preds_ner += preds_ner
        p, r, f1 = score_ner.score_by_entity(dev_preds_ner, dev_gold_tags)
        for i in range(len(dev_batch_ner)):
            assert len(dev_preds_ner[i]) == len(dev_gold_tags[i])

        train_loss = train_loss / len(train_batch_pos)  # avg loss per batch
        train_loss_dep = train_loss_dep / len(train_batch_pos)
        train_loss_pos = train_loss_pos / len(train_batch_pos)
        train_loss_ner = train_loss_ner / len(train_batch_pos)

        print(
            "step {}: train_loss = {:.6f}, train_loss_dep = {:.6f}, train_loss_pos = {:.6f}, train_loss_ner = {:.6f}, dev_las_score = {:.4f}, dev_uas_score = {:.4f}, dev_pos = {:.4f}, dev_ner_p = {:.4f}, dev_ner_r = {:.4f}, dev_ner_f1 = {:.4f} ".format(
                global_step,
                train_loss,
                train_loss_dep,
                train_loss_pos,
                train_loss_ner,
                las_dev,
                uas_dev,
                accuracy_pos_dev,
                p,
                r,
                f1,
            )
        )

        # save best model
        if las_dev + accuracy_pos_dev + f1 >= (las_score_history + upos_score_history + f1_score_history):
            las_score_history = las_dev
            upos_score_history = accuracy_pos_dev
            uas_score_history = uas_dev
            f1_score_history = f1
            trainer.save(model_file)
            print("new best model saved.")
        train_loss = 0
        train_loss_pos = 0
        train_loss_dep = 0
        train_loss_ner = 0

        print("")
        train_batch_dep.reshuffle()
        train_batch_pos.reshuffle()
        train_batch_ner.reshuffle()

    print("Training ended with {} epochs.".format(epoch))

    best_las, uas, upos, f1 = (
        las_score_history * 100,
        uas_score_history * 100,
        upos_score_history * 100,
        f1_score_history * 100,
    )
    print("Best dev las = {:.2f}, uas = {:.2f}, upos = {:.2f}, f1 = {:.2f}".format(best_las, uas, upos, f1))


def evaluate(args):
    # file paths
    system_pred_file = args["output_file_dep"]
    gold_file = args["eval_file_dep"]
    model_file = args["save_dir"] + "/" + "phonlp.pt"

    checkpoint = torch.load(model_file, lambda storage, loc: storage)
    loaded_args = checkpoint["config"]
    vocab = MultiVocab.load_state_dict(checkpoint["vocab"])
    config_phobert = AutoConfig.from_pretrained(loaded_args["pretrained_lm"], output_hidden_states=True)
    tokenizer = AutoTokenizer.from_pretrained(loaded_args["pretrained_lm"], use_fast=False)

    # load model
    print("Loading model from: {}".format(model_file))
    use_cuda = args["cuda"] and not args["cpu"]
    trainer = JointTrainer(model_file=model_file, use_cuda=use_cuda, config_phobert=config_phobert)

    # load data
    print("Loading data with batch size {}...".format(args["batch_size"]))

    test_doc_dep = Document(CoNLL.conll2dict(input_file=args["eval_file_dep"]))

    test_batch_pos = DataLoaderPOS(
        args["eval_file_pos"],
        args["batch_size"],
        args,
        vocab=vocab,
        sort_during_eval=True,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    test_batch_dep = DataLoaderDep(
        test_doc_dep,
        args["batch_size"],
        args,
        vocab=vocab,
        sort_during_eval=True,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )
    test_batch_ner = DataLoaderNER(
        args["eval_file_ner"],
        args["batch_size"],
        args,
        vocab=vocab,
        evaluation=True,
        tokenizer=tokenizer,
        max_seq_length=args["max_sequence_length"],
    )

    print("Start evaluation...")
    test_preds_dep = []
    test_preds_upos = []
    test_preds_ner = []
    for batch in test_batch_dep:
        preds_dep = trainer.predict_dep(batch)
        test_preds_dep += preds_dep
    test_preds_dep = util.unsort(test_preds_dep, test_batch_dep.data_orig_idx_dep)
    test_batch_dep.doc_dep.set([HEAD, DEPREL], [y for x in test_preds_dep for y in x])
    CoNLL.dict2conll(test_batch_dep.doc_dep.to_dict(), system_pred_file)
    _, _, las, uas = score_dep.score(system_pred_file, gold_file)

    for batch in test_batch_pos:
        preds_pos = trainer.predict_pos(batch)
        test_preds_upos += preds_pos
    test_preds_upos = util.unsort(test_preds_upos, test_batch_pos.data_orig_idx_pos)
    accuracy_pos = score_pos.score_acc(test_preds_upos, test_batch_pos.upos)

    for batch in test_batch_ner:
        preds_ner = trainer.predict_ner(batch)
        test_preds_ner += preds_ner
    p, r, f1 = score_ner.score_by_entity(test_preds_ner, test_batch_ner.tags)

    print(
        "{} POS tagging: {:.2f}, NER: {:.2f}, Dependency parsing: {:.2f}/{:.2f}".format(
            "Evaluation results: ", accuracy_pos * 100, f1 * 100, las * 100, uas * 100
        )
    )


def annotate(input_file=None, output_file=None, args=None, batch_size=1):
    model_file = args["save_dir"] + "/" + "phonlp.pt"
    print("Loading model from: {}".format(model_file))
    checkpoint = torch.load(model_file, lambda storage, loc: storage)
    args = checkpoint["config"]
    vocab = MultiVocab.load_state_dict(checkpoint["vocab"])
    # load model
    tokenizer = AutoTokenizer.from_pretrained(args["pretrained_lm"], use_fast=False)
    config_phobert = AutoConfig.from_pretrained(args["pretrained_lm"], output_hidden_states=True)
    model = JointModel(args, vocab, config_phobert, tokenizer)
    model.load_state_dict(checkpoint["model"], strict=False)
    if torch.cuda.is_available() is False:
        model.to(torch.device("cpu"))
    else:
        model.to(torch.device("cuda"))
    model.eval()
    model.annotate(input_file=input_file, output_file=output_file, batch_size=batch_size)


if __name__ == "__main__":
    main()



import logging
import random

import torch
from phonlp.models.common.data import get_long_tensor, sort_all
from phonlp.models.common.doc import DEPREL, HEAD, TEXT
from phonlp.models.common.vocab import PAD_ID, ROOT_ID
from phonlp.models.ner.vocab import TagVocab
from phonlp.models.pos.vocab import CharVocab, MultiVocab, WordVocab


logger = logging.getLogger("PhoNLPToolkit")


class DataLoaderPOS:
    def __init__(
        self,
        path_file,
        batch_size,
        args,
        vocab=None,
        evaluation=False,
        sort_during_eval=False,
        tokenizer=None,
        max_seq_length=None,
    ):
        self.batch_size = batch_size
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval
        self.sort_during_eval = sort_during_eval

        data = read_file(path_file)
        self.upos = [[w[1] for w in sent] for sent in data]
        # handle vocab
        self.vocab = vocab

        data = self.preprocess(data, self.vocab, tokenizer, max_seq_length)
        if evaluation is True:
            print("Number of evaluation sentences for POS tagging: ", len(data))
        else:
            print("Number of training sentences for POS tagging: ", len(data))
        # shuffle for training
        if self.shuffled:
            random.shuffle(data)
        self.num_examples = len(data)

        # chunk into batches
        self.data = self.chunk_batches(data)

    def preprocess(self, data, vocab, tokenizer, max_seq_length):
        pad_id = 1
        cls_id = 0
        sep_id = 2
        processed = []
        for sent in data:
            input_ids = [cls_id]
            firstSWindices = [len(input_ids)]
            for w in sent:
                word_token = tokenizer.encode(w[0])
                input_ids += word_token[1 : (len(word_token) - 1)]
                firstSWindices.append(len(input_ids))
            firstSWindices = firstSWindices[: (len(firstSWindices) - 1)]
            input_ids.append(sep_id)
            if len(input_ids) > max_seq_length:
                input_ids = input_ids[:max_seq_length]
            else:
                input_ids = (
                    input_ids
                    + [
                        pad_id,
                    ]
                    * (max_seq_length - len(input_ids))
                )

            processed_sent = [input_ids]
            processed_sent += [firstSWindices]
            processed_sent += [vocab["upos"].map([w[1] for w in sent])]
            processed.append(processed_sent)
        return processed

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        """ Get a batch with index. """
        if not isinstance(key, int):
            raise TypeError
        if key > 0 and key % len(self.data) == 0:
            self.reshuffle()
        batch_key = key % len(self.data)
        batch = self.data[batch_key]
        batch_size = len(batch)
        batch = list(zip(*batch))
        assert len(batch) == 3

        lens = [len(x) for x in batch[2]]
        batch, orig_idx = sort_all(batch, lens)

        # convert to tensors
        tokens_phobert = batch[0]
        tokens_phobert = get_long_tensor(tokens_phobert, batch_size, pad_id=1)
        first_subword = batch[1]
        first_subword = get_long_tensor(first_subword, batch_size)
        upos = get_long_tensor(batch[2], batch_size)
        sentlens = [len(x) for x in batch[1]]
        return tokens_phobert, first_subword, upos, orig_idx, sentlens

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def resolve_none(self, data):
        # replace None to '_'
        for sent_idx in range(len(data)):
            for tok_idx in range(len(data[sent_idx])):
                for feat_idx in range(len(data[sent_idx][tok_idx])):
                    if data[sent_idx][tok_idx][feat_idx] is None:
                        data[sent_idx][tok_idx][feat_idx] = "_"
        return data

    def reshuffle(self):
        data = [y for x in self.data for y in x]
        random.shuffle(self.data)
        self.data = self.chunk_batches(data)

    def chunk_batches(self, data):
        res = []

        if not self.eval:
            # sort sentences (roughly) by length for better memory utilization
            data = sorted(data, key=lambda x: len(x[2]), reverse=random.random() > 0.5)
        elif self.sort_during_eval:
            (data,), self.data_orig_idx_pos = sort_all([data], [len(x[2]) for x in data])

        current = []
        for x in data:
            if len(current) == self.batch_size:
                res.append(current)
                current = []
            current.append(x)

        if len(current) > 0:
            res.append(current)

        return res


class DataLoaderDep:
    def __init__(
        self,
        doc_dep,
        batch_size,
        args,
        vocab=None,
        evaluation=False,
        sort_during_eval=False,
        tokenizer=None,
        max_seq_length=None,
    ):
        self.batch_size = batch_size
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval
        self.sort_during_eval = sort_during_eval
        self.doc_dep = doc_dep
        data_dep = self.load_doc(doc_dep)  # list sentences of list words of list fields
        self.head = [[w[1] for w in sent] for sent in data_dep]
        self.deprel = [[w[2] for w in sent] for sent in data_dep]

        # handle vocab
        self.vocab = vocab

        data_dep = self.preprocess(data_dep, self.vocab, tokenizer, max_seq_length)
        if evaluation is True:
            print("Number of evaluation sentences for dependency parsing: ", len(data_dep))
        else:
            print("Number of training sentences for dependency parsing: ", len(data_dep))

        # shuffle for training
        if self.shuffled:
            random.shuffle(data_dep)
        # chunk into batches
        self.data_dep = self.chunk_batches(data_dep)

    def preprocess(self, data_dep, vocab, tokenizer, max_seq_length):
        pad_id = 1
        cls_id = 0
        sep_id = 2
        processed_dep = []
        for sent in data_dep:
            input_ids = [cls_id]
            firstSWindices = [len(input_ids)]
            root_token = tokenizer.encode("[ROOT]")
            input_ids += root_token[1 : (len(root_token) - 1)]
            firstSWindices.append(len(input_ids))
            for w in sent:
                word_token = tokenizer.encode(w[0])
                input_ids += word_token[1 : (len(word_token) - 1)]
                firstSWindices.append(len(input_ids))
            firstSWindices = firstSWindices[: (len(firstSWindices) - 1)]
            input_ids.append(sep_id)
            if len(input_ids) > max_seq_length:
                input_ids = input_ids[:max_seq_length]
            else:
                input_ids = (
                    input_ids
                    + [
                        pad_id,
                    ]
                    * (max_seq_length - len(input_ids))
                )

            processed_sent = [input_ids]
            processed_sent += [firstSWindices]
            processed_sent += [[ROOT_ID] + vocab["word"].map([w[0] for w in sent])]
            ##
            processed_sent += [[[ROOT_ID]] + [vocab["char"].map([x for x in w[0]]) for w in sent]]
            processed_sent += [[to_int(w[1], ignore_error=self.eval) for w in sent]]
            processed_sent += [vocab["deprel"].map([w[2] for w in sent])]
            processed_dep.append(processed_sent)

        return processed_dep  # [sent1,sent2,...] sent1 = list of list

    def __len__(self):
        return len(self.data_dep)

    def __getitem__(self, key):
        """ Get a batch with index. """
        if not isinstance(key, int):
            raise TypeError
        if key > 0 and key % len(self.data_dep) == 0:
            self.reshuffle()
        dep_key = key % len(self.data_dep)
        dep_batch = self.data_dep[dep_key]
        dep_batch_size = len(dep_batch)
        dep_batch = list(zip(*dep_batch))

        assert len(dep_batch) == 6

        dep_lens = [len(x) for x in dep_batch[2]]
        dep_batch, dep_orig_idx = sort_all(dep_batch, dep_lens)

        dep_batch_words = [w for sent in dep_batch[3] for w in sent]
        dep_word_lens = [len(x) for x in dep_batch_words]
        dep_batch_words, dep_word_orig_idx = sort_all([dep_batch_words], dep_word_lens)
        dep_batch_words = dep_batch_words[0]  # [word1,...], word1 = list of tokens
        dep_word_lens = [len(x) for x in dep_batch_words]
        dep_wordchars = get_long_tensor(dep_batch_words, len(dep_word_lens))
        dep_number_of_words = dep_wordchars.size(0)
        dep_words = dep_batch[2]
        dep_words = get_long_tensor(dep_words, dep_batch_size)
        dep_words_mask = torch.eq(dep_words, PAD_ID)

        # convert to tensors
        dep_tokens_phobert = dep_batch[0]
        dep_tokens_phobert = get_long_tensor(dep_tokens_phobert, dep_batch_size, pad_id=1)
        dep_first_subword = dep_batch[1]
        dep_first_subword = get_long_tensor(dep_first_subword, dep_batch_size)
        dep_sentlens = [len(x) for x in dep_batch[1]]
        dep_head = get_long_tensor(dep_batch[4], dep_batch_size)
        dep_deprel = get_long_tensor(dep_batch[5], dep_batch_size)
        dep_data = (
            dep_tokens_phobert,
            dep_first_subword,
            dep_words_mask,
            dep_head,
            dep_deprel,
            dep_number_of_words,
            dep_orig_idx,
            dep_sentlens,
        )
        return dep_data

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def reshuffle(self):
        data_dep = [y for x in self.data_dep for y in x]
        random.shuffle(self.data_dep)
        self.data_dep = self.chunk_batches(data_dep)
        random.shuffle(self.data_dep)

    def load_doc(self, doc):
        data = doc.get([TEXT, HEAD, DEPREL], as_sentences=True)  # [[[cac value trong 1 token]]]
        data = self.resolve_none(data)
        return data

    def resolve_none(self, data):
        # replace None to '_'
        for sent_idx in range(len(data)):
            for tok_idx in range(len(data[sent_idx])):
                for feat_idx in range(len(data[sent_idx][tok_idx])):
                    if data[sent_idx][tok_idx][feat_idx] is None:
                        data[sent_idx][tok_idx][feat_idx] = "_"
        return data

    def chunk_batches(self, data):
        res = []
        if not self.eval:
            # sort sentences (roughly) by length for better memory utilization
            data = sorted(data, key=lambda x: len(x[2]), reverse=random.random() > 0.5)
        elif self.sort_during_eval:
            (data,), self.data_orig_idx_dep = sort_all([data], [len(x[2]) for x in data])
        current = []
        for x in data:
            if len(current) >= self.batch_size:
                res.append(current)
                current = []
            current.append(x)

        if len(current) > 0:
            res.append(current)
        return res  # list of list by sentences of list by token


class DataLoaderNER:
    def __init__(self, path_file, batch_size, args, vocab=None, evaluation=False, tokenizer=None, max_seq_length=None):
        self.batch_size = batch_size
        self.args = args
        self.eval = evaluation
        self.shuffled = not self.eval

        data = read_file(path_file)
        self.tags = [[w[1] for w in sent] for sent in data]
        self.vocab = vocab

        data = self.preprocess(data, self.vocab, args, tokenizer, max_seq_length)
        if evaluation is True:
            print("Number of evaluation sentences for NER: ", len(data))
        else:
            print("Number of training sentences for NER: ", len(data))
        # shuffle for training
        if self.shuffled:
            random.shuffle(data)
        self.num_examples = len(data)

        # chunk into batches
        self.data = self.chunk_batches(data)

    def preprocess(self, data, vocab, args, tokenizer, max_seq_length):
        processed = []
        sep_id = 2
        cls_id = 0
        pad_id = 1
        for sent in data:
            input_ids = [cls_id]
            firstSWindices = [len(input_ids)]
            for w in sent:
                word_token = tokenizer.encode(w[0])
                input_ids += word_token[1 : (len(word_token) - 1)]
                firstSWindices.append(len(input_ids))
            firstSWindices = firstSWindices[: (len(firstSWindices) - 1)]
            input_ids.append(sep_id)
            if len(input_ids) > max_seq_length:
                input_ids = input_ids[:max_seq_length]
            else:
                input_ids = (
                    input_ids
                    + [
                        pad_id,
                    ]
                    * (max_seq_length - len(input_ids))
                )
            processed_sent = [input_ids]
            processed_sent += [firstSWindices]
            processed_sent += [vocab["word"].map([w[0] for w in sent])]
            processed_sent += [vocab["ner_tag"].map([w[1] for w in sent])]
            processed.append(processed_sent)
        return processed

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        """ Get a batch with index. """
        if not isinstance(key, int):
            raise TypeError
        if key < 0:  # or key % len(self.data):
            raise IndexError
        key = key % len(self.data)
        batch = self.data[key]
        batch_size = len(batch)
        batch = list(zip(*batch))
        assert len(batch) == 4

        # sort sentences by lens for easy RNN operations
        sentlens = [len(x) for x in batch[2]]
        batch, orig_idx = sort_all(batch, sentlens)
        sentlens = [len(x) for x in batch[2]]

        words = get_long_tensor(batch[2], batch_size)
        words_mask = torch.eq(words, PAD_ID)
        # convert to tensors
        tokens_phobert = batch[0]
        tokens_phobert = get_long_tensor(tokens_phobert, batch_size, pad_id=1)
        first_subword = batch[1]
        first_subword = get_long_tensor(first_subword, batch_size)
        tags = get_long_tensor(batch[3], batch_size)
        return tokens_phobert, first_subword, words_mask, tags, orig_idx, sentlens

    def __iter__(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)

    def reshuffle(self):
        data = [y for x in self.data for y in x]
        random.shuffle(data)
        self.data = self.chunk_batches(data)

    def chunk_batches(self, data):
        data = [data[i : i + self.batch_size] for i in range(0, len(data), self.batch_size)]
        return data


def read_file(file):
    f = open(file)
    doc, sent = [], []
    for line in f:
        line = line.strip()
        if len(line) == 0:
            if len(sent) > 0:
                doc.append(sent)
                sent = []
        else:
            array = line.split("\t")
            sent += [array]
    if len(sent) > 0:
        doc.append(sent)
    return doc


class BuildVocab:
    def __init__(self, args, path_pos, doc_dep, path_ner):
        self.args = args
        data_pos = read_file(path_pos)
        data_dep = self.load_doc_dep(doc_dep)
        data_ner = read_file(path_ner)
        self.vocab = self.build_vocab(data_dep, data_pos, data_ner)

    def build_vocab(self, data_dep, data_pos, data_ner):
        data = data_dep + data_pos + data_ner
        uposvocab = WordVocab(data_pos, idx=1)
        deprelvocab = WordVocab(data_dep, idx=2)
        ner_tag = TagVocab(data_ner, idx=1)
        charvocab = CharVocab(data)
        wordvocab = WordVocab(data, cutoff=7, lower=True)
        vocab = MultiVocab(
            {"upos": uposvocab, "deprel": deprelvocab, "ner_tag": ner_tag, "char": charvocab, "word": wordvocab}
        )
        return vocab

    def load_doc_dep(self, doc):
        data = doc.get([TEXT, HEAD, DEPREL], as_sentences=True)
        data = self.resolve_none(data)
        return data

    def resolve_none(self, data):
        # replace None to '_'
        for sent_idx in range(len(data)):
            for tok_idx in range(len(data[sent_idx])):
                for feat_idx in range(len(data[sent_idx][tok_idx])):
                    if data[sent_idx][tok_idx][feat_idx] is None:
                        data[sent_idx][tok_idx][feat_idx] = "_"
        return data


def to_int(string, ignore_error=False):
    try:
        res = int(string)
    except ValueError as err:
        if ignore_error:
            return 0
        else:
            raise err
    return res



# -*- coding: utf-8 -*-
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from phonlp.models.common import utils as util
from phonlp.models.common.biaffine import DeepBiaffineScorer
from phonlp.models.common.chuliu_edmonds import chuliu_edmonds_one_root
from phonlp.models.common.crf import CRFLoss, viterbi_decode
from phonlp.models.common.data import get_long_tensor, sort_all
from phonlp.models.common.dropout import LockedDropout, WordDropout
from phonlp.models.common.vocab import PAD_ID, ROOT_ID
from torch.nn.utils.rnn import PackedSequence, pack_padded_sequence, pad_packed_sequence
from tqdm import tqdm
from transformers import AutoModel, BertPreTrainedModel


class JointModel(BertPreTrainedModel):
    def __init__(self, args, vocab, config, tokenizer):
        super(JointModel, self).__init__(config)

        self.vocab = vocab
        self.args = args
        self.unsaved_modules = []
        self.config = config
        self.tokenizer = tokenizer

        # input layers
        self.input_size = 0

        self.phobert = AutoModel.from_config(self.config)
        self.input_size += self.config.to_dict()["hidden_size"]

        self.drop_replacement_ner = nn.Parameter(
            torch.randn(self.input_size + self.args["tag_emb_dim"])
            / np.sqrt(self.input_size + self.args["tag_emb_dim"])
        )
        self.drop_replacement_dep = nn.Parameter(
            torch.randn(self.input_size + self.args["tag_emb_dim"])
            / np.sqrt(self.input_size + self.args["tag_emb_dim"])
        )
        self.drop_replacement_pos = nn.Parameter(torch.randn(self.input_size) / np.sqrt(self.input_size))

        self.upos_hid = nn.Linear(self.input_size, self.args["deep_biaff_hidden_dim"])
        self.upos_clf = nn.Linear(self.args["deep_biaff_hidden_dim"], len(vocab["upos"]))

        self.upos_emb_matrix_ner = nn.Parameter(
            torch.rand(len(vocab["upos"]), self.args["tag_emb_dim"]), requires_grad=True
        )
        self.upos_emb_matrix_dep = nn.Parameter(
            torch.rand(len(vocab["upos"]), self.args["tag_emb_dim"]), requires_grad=True
        )
        self.upos_clf.weight.data.zero_()
        self.upos_clf.bias.data.zero_()

        self.dep_hid = nn.Linear(
            self.input_size + self.args["tag_emb_dim"], self.input_size + self.args["tag_emb_dim"]
        )

        # classifiers
        self.unlabeled = DeepBiaffineScorer(
            self.input_size + self.args["tag_emb_dim"],
            self.input_size + self.args["tag_emb_dim"],
            self.args["deep_biaff_hidden_dim"],
            1,
            pairwise=True,
            dropout=args["dropout"],
        )
        self.deprel = DeepBiaffineScorer(
            self.input_size + self.args["tag_emb_dim"],
            self.input_size + self.args["tag_emb_dim"],
            self.args["deep_biaff_hidden_dim"],
            len(vocab["deprel"]),
            pairwise=True,
            dropout=args["dropout"],
        )
        if args["linearization"]:
            self.linearization = DeepBiaffineScorer(
                self.input_size + self.args["tag_emb_dim"],
                self.input_size + self.args["tag_emb_dim"],
                self.args["deep_biaff_hidden_dim"],
                1,
                pairwise=True,
                dropout=args["dropout"],
            )
        if args["distance"]:
            self.distance = DeepBiaffineScorer(
                self.input_size + self.args["tag_emb_dim"],
                self.input_size + self.args["tag_emb_dim"],
                self.args["deep_biaff_hidden_dim"],
                1,
                pairwise=True,
                dropout=args["dropout"],
            )

        self.ner_tag_clf = nn.Linear(self.input_size + self.args["tag_emb_dim"], len(self.vocab["ner_tag"]))
        self.ner_tag_clf.bias.data.zero_()

        # criterion
        self.crit_ner = CRFLoss(len(self.vocab["ner_tag"]))
        self.crit_dep = nn.CrossEntropyLoss(ignore_index=-1, reduction="sum")  # ignore padding
        self.crit_pos = nn.CrossEntropyLoss(ignore_index=1)
        self.drop_ner = nn.Dropout(args["dropout"])
        self.worddrop_ner = WordDropout(args["word_dropout"])
        self.lockeddrop = LockedDropout(args["word_dropout"])

        self.drop_pos = nn.Dropout(args["dropout"])
        self.worddrop_pos = WordDropout(args["word_dropout"])

        self.drop_dep = nn.Dropout(args["dropout"])
        self.worddrop_dep = WordDropout(args["word_dropout"])

    def tagger_forward(self, tokens_phobert, words_phobert, sentlens):
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs = []

        phobert_emb = self.phobert(tokens_phobert)[2][-1]
        phobert_emb = torch.cat(
            [torch.index_select(phobert_emb[i], 0, words_phobert[i]).unsqueeze(0) for i in range(phobert_emb.size(0))],
            dim=0,
        )
        if torch.cuda.is_available():
            phobert_emb = phobert_emb.cuda()
        phobert_emb = pack(phobert_emb)
        inputs += [phobert_emb]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, phobert_emb.batch_sizes), batch_first=True)[0]

        inputs_pos = inputs[0].data
        inputs_pos = self.worddrop_pos(inputs_pos, self.drop_replacement_pos)

        upos_hid = F.relu(self.upos_hid(inputs_pos))
        upos_pred = self.upos_clf(self.drop_pos(upos_hid))

        preds_pos = [pad(upos_pred).max(2)[1]]

        pos_dis = F.softmax(pad(upos_pred), dim=-1)
        upos_embed_matrix_dup = self.upos_emb_matrix_ner.repeat(pos_dis.size(0), 1, 1)
        pos_emb = torch.matmul(pos_dis, upos_embed_matrix_dup)
        pos_emb = pack(pos_emb)
        inputs += [pos_emb]

        inputs = torch.cat([x.data for x in inputs], 1)
        inputs = self.worddrop_ner(inputs, self.drop_replacement_ner)

        ner_pred = self.ner_tag_clf(inputs)
        logits = pad(F.relu(ner_pred)).contiguous()

        return preds_pos, logits

    def dep_forward(self, tokens_phobert, first_subword, sentlens):
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs = []
        phobert_emb = self.phobert(tokens_phobert)[2][-1]
        phobert_emb = torch.cat(
            [torch.index_select(phobert_emb[i], 0, first_subword[i]).unsqueeze(0) for i in range(phobert_emb.size(0))],
            dim=0,
        )
        if torch.cuda.is_available():
            phobert_emb = phobert_emb.cuda()
        phobert_emb = pack(phobert_emb)
        inputs += [phobert_emb]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, phobert_emb.batch_sizes), batch_first=True)[0]

        inputs_pos = inputs[0].data
        inputs_pos = self.worddrop_pos(inputs_pos, self.drop_replacement_pos)

        upos_hid = F.relu(self.upos_hid(inputs_pos))
        upos_pred = self.upos_clf(self.drop_pos(upos_hid))
        pos_dis = F.softmax(pad(upos_pred), dim=-1)
        upos_embed_matrix_dup = self.upos_emb_matrix_dep.repeat(pos_dis.size(0), 1, 1)
        pos_emb = torch.matmul(pos_dis, upos_embed_matrix_dup)
        pos_emb = pack(pos_emb)
        inputs += [pos_emb]

        inputs = torch.cat([x.data for x in inputs], 1)
        inputs = self.worddrop_dep(inputs, self.drop_replacement_dep)

        hidden_out = self.dep_hid(inputs)
        hidden_out = pad(hidden_out)

        unlabeled_scores = self.unlabeled(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
        deprel_scores = self.deprel(self.drop_dep(hidden_out), self.drop_dep(hidden_out))

        if self.args["linearization"] or self.args["distance"]:
            head_offset = torch.arange(first_subword.size(1), device=unlabeled_scores.device).view(1, 1, -1).expand(
                first_subword.size(0), -1, -1
            ) - torch.arange(first_subword.size(1), device=unlabeled_scores.device).view(1, -1, 1).expand(
                first_subword.size(0), -1, -1
            )

        if self.args["linearization"]:
            lin_scores = self.linearization(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
            unlabeled_scores += F.logsigmoid(lin_scores * torch.sign(head_offset).float()).detach()

        if self.args["distance"]:
            dist_scores = self.distance(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
            dist_pred = 1 + F.softplus(dist_scores)
            dist_target = torch.abs(head_offset)
            dist_kld = -torch.log((dist_target.float() - dist_pred) ** 2 / 2 + 1)
            unlabeled_scores += dist_kld.detach()

        diag = torch.eye(unlabeled_scores.size(-1), dtype=torch.bool, device=unlabeled_scores.device).unsqueeze(0)
        unlabeled_scores.masked_fill_(diag, -float("inf"))

        preds = []
        preds.append(F.log_softmax(unlabeled_scores, 2).detach().cpu().numpy())
        preds.append(deprel_scores.max(3)[1].detach().cpu().numpy())
        return preds

    def annotate(self, text=None, input_file=None, output_file=None, batch_size=1, output_type=""):
        if text is not None:
            data = [text.split(" ")]
        else:
            f = open(input_file)
            data = []
            for line in f:
                line = line.strip()
                if len(line) != 0:
                    data.append(line.split(" "))
            f.close()
            print("The number of sentences: ", len(data))
        data_tagger = self.process_data_tagger(batch_text=data)
        data_parser = self.process_data_parser(batch_text=data)
        data_parser = self.chunk_batches(data_parser, batch_size)
        data_tagger = self.chunk_batches(data_tagger, batch_size)
        test_preds_pos = []
        test_preds_dep = []
        test_preds_ner = []
        for i in tqdm(range(len(data_tagger))):
            tokens_phobert, first_subword, words_mask, number_of_words, orig_idx, sentlens = self.get_batch(
                i, data_tagger
            )
            tokens_phobert1, first_subword1, words_mask1, number_of_words1, orig_idx1, sentlens1 = self.get_batch(
                i, data_parser
            )
            if torch.cuda.is_available():
                tokens_phobert, first_subword, words_mask = (
                    tokens_phobert.cuda(),
                    first_subword.cuda(),
                    words_mask.cuda(),
                )
                tokens_phobert1, first_subword1, words_mask1 = (
                    tokens_phobert1.cuda(),
                    first_subword1.cuda(),
                    words_mask1.cuda(),
                )

            preds_dep = self.dep_forward(tokens_phobert1, first_subword1, sentlens1)
            preds_pos, logits = self.tagger_forward(tokens_phobert, first_subword, sentlens)
            batch_size = tokens_phobert.size(0)
            # DEP
            head_seqs = [
                chuliu_edmonds_one_root(adj[:l, :l])[1:] for adj, l in zip(preds_dep[0], sentlens1)
            ]  # remove attachment for the root
            deprel_seqs = [
                self.vocab["deprel"].unmap([preds_dep[1][i][j + 1][h] for j, h in enumerate(hs)])
                for i, hs in enumerate(head_seqs)
            ]
            pred_tokens = [
                [[str(head_seqs[i][j]), deprel_seqs[i][j]] for j in range(sentlens1[i] - 1)] for i in range(batch_size)
            ]
            pred_tokens_dep = util.unsort(pred_tokens, orig_idx1)

            # POS
            upos_seqs = [self.vocab["upos"].unmap(sent) for sent in preds_pos[0]]
            pred_tokens_pos = [
                [[upos_seqs[i][j]] for j in range(sentlens[i])] for i in range(batch_size)
            ]  # , xpos_seqs[i][j], feats_seqs[i][j]
            pred_tokens_pos = util.unsort(pred_tokens_pos, orig_idx)

            trans = self.crit_ner._transitions.data.cpu().numpy()
            scores = logits.data.cpu().numpy()
            bs = logits.size(0)
            tag_seqs = []
            for i in range(bs):
                tags, _ = viterbi_decode(scores[i, : sentlens[i]], trans)
                tags = self.vocab["ner_tag"].unmap(tags)
                tag_seqs += [tags]
            tag_seqs = util.unsort(tag_seqs, orig_idx)
            test_preds_ner += tag_seqs
            test_preds_dep += pred_tokens_dep
            test_preds_pos += pred_tokens_pos
        test_preds_dep = util.unsort(test_preds_dep, self.data_orig_idx)
        test_preds_pos = util.unsort(test_preds_pos, self.data_orig_idx)
        test_preds_ner = util.unsort(test_preds_ner, self.data_orig_idx)
        if text is not None:
            return (data, test_preds_pos, test_preds_ner, test_preds_dep)
        else:
            f = open(output_file, "w")
            for i in range(len(data)):
                for j in range(len(data[i])):
                    if output_type == "conll":
                        f.write(
                            str(j + 1)
                            + "\t"
                            + data[i][j]
                            + "\t"
                            + "_"
                            + "\t"
                            + "_"
                            + "\t"
                            + test_preds_pos[i][j][0]
                            + "\t"
                            + "_"
                            + "\t"
                            + test_preds_dep[i][j][0]
                            + "\t"
                            + test_preds_dep[i][j][1]
                            + "\t"
                            + "_"
                            + "\t"
                            + test_preds_ner[i][j]
                            + "\n"
                        )
                    else:
                        f.write(
                            str(j + 1)
                            + "\t"
                            + data[i][j]
                            + "\t"
                            + test_preds_pos[i][j][0]
                            + "\t"
                            + test_preds_ner[i][j]
                            + "\t"
                            + test_preds_dep[i][j][0]
                            + "\t"
                            + test_preds_dep[i][j][1]
                            + "\n"
                        )
                f.write("\n")
            f.close()

    def print_out(self, output, output_type=""):
        data, test_preds_pos, test_preds_ner, test_preds_dep = output
        for i in range(len(data)):
            for j in range(len(data[i])):
                if output_type == "conll":
                    print(
                        str(j + 1)
                        + "\t"
                        + data[i][j]
                        + "\t"
                        + "_"
                        + "\t"
                        + "_"
                        + "\t"
                        + test_preds_pos[i][j][0]
                        + "\t"
                        + "_"
                        + "\t"
                        + test_preds_dep[i][j][0]
                        + "\t"
                        + test_preds_dep[i][j][1]
                        + "\t"
                        + "_"
                        + "\t"
                        + test_preds_ner[i][j]
                    )
                else:
                    print(
                        str(j + 1)
                        + "\t"
                        + data[i][j]
                        + "\t"
                        + test_preds_pos[i][j][0]
                        + "\t"
                        + test_preds_ner[i][j]
                        + "\t"
                        + test_preds_dep[i][j][0]
                        + "\t"
                        + test_preds_dep[i][j][1]
                    )

    def process_data_tagger(self, batch_text):
        cls_id = 0
        sep_id = 2
        processed = []
        for sent in batch_text:
            input_ids = [cls_id]
            firstSWindices = [len(input_ids)]
            for w in sent:
                word_token = self.tokenizer.encode(w)
                input_ids += word_token[1 : (len(word_token) - 1)]
                firstSWindices.append(len(input_ids))
            firstSWindices = firstSWindices[: (len(firstSWindices) - 1)]
            input_ids.append(sep_id)
            processed_sent = [input_ids]
            processed_sent += [firstSWindices]
            processed_sent += [self.vocab["word"].map([w for w in sent])]
            processed_sent += [[self.vocab["char"].map([x for x in w]) for w in sent]]
            processed.append(processed_sent)
        return processed

    def process_data_parser(self, batch_text):
        cls_id = 0
        sep_id = 2
        processed = []
        for sent in batch_text:
            input_ids = [cls_id]
            firstSWindices = [len(input_ids)]
            root_token = self.tokenizer.encode("[ROOT]")
            input_ids += root_token[1 : (len(root_token) - 1)]
            firstSWindices.append(len(input_ids))
            for w in sent:
                word_token = self.tokenizer.encode(w)
                input_ids += word_token[1 : (len(word_token) - 1)]
                firstSWindices.append(len(input_ids))
            firstSWindices = firstSWindices[: (len(firstSWindices) - 1)]
            input_ids.append(sep_id)

            processed_sent = [input_ids]
            processed_sent += [firstSWindices]
            processed_sent += [[ROOT_ID] + self.vocab["word"].map([w for w in sent])]
            processed_sent += [[[ROOT_ID]] + [self.vocab["char"].map([x for x in w]) for w in sent]]
            processed.append(processed_sent)
        return processed

    def chunk_batches(self, data, batch_size):
        res = []
        (data,), self.data_orig_idx = sort_all([data], [len(x[2]) for x in data])
        current = []
        for x in data:
            if len(current) >= batch_size:
                res.append(current)
                current = []
            current.append(x)

        if len(current) > 0:
            res.append(current)
        return res

    def get_batch(self, key, data_chunk):
        batch = data_chunk[key]
        batch_size = len(batch)
        batch = list(zip(*batch))
        assert len(batch) == 4
        # print(batch)
        lens = [len(x) for x in batch[2]]
        batch, orig_idx = sort_all(batch, lens)

        batch_words = [w for sent in batch[3] for w in sent]
        word_lens = [len(x) for x in batch_words]
        batch_words, word_orig_idx = sort_all([batch_words], word_lens)
        batch_words = batch_words[0]  # [word1,...], word1 = list of tokens
        word_lens = [len(x) for x in batch_words]
        wordchars = get_long_tensor(batch_words, len(word_lens))
        number_of_words = wordchars.size(0)
        words = batch[2]
        words = get_long_tensor(words, batch_size)
        words_mask = torch.eq(words, PAD_ID)

        # convert to tensors
        tokens_phobert = batch[0]
        tokens_phobert = get_long_tensor(tokens_phobert, batch_size, pad_id=1)
        first_subword = batch[1]
        first_subword = get_long_tensor(first_subword, batch_size)
        sentlens = [len(x) for x in batch[1]]
        data = (tokens_phobert, first_subword, words_mask, number_of_words, orig_idx, sentlens)
        return data



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from phonlp.models.common.biaffine import DeepBiaffineScorer
from phonlp.models.common.crf import CRFLoss
from phonlp.models.common.dropout import LockedDropout, WordDropout
from torch.nn.utils.rnn import PackedSequence, pack_padded_sequence, pad_packed_sequence
from transformers import AutoModel, BertPreTrainedModel


class JointModel(BertPreTrainedModel):
    def __init__(self, args, vocab, config):
        super(JointModel, self).__init__(config)

        self.vocab = vocab
        self.args = args
        self.unsaved_modules = []
        self.config = config

        # input layers
        self.input_size = 0

        self.phobert = AutoModel.from_pretrained(args["pretrained_lm"], config=self.config)
        self.input_size += self.config.to_dict()["hidden_size"]

        self.drop_replacement_ner = nn.Parameter(
            torch.randn(self.input_size + self.args["tag_emb_dim"])
            / np.sqrt(self.input_size + self.args["tag_emb_dim"])
        )
        self.drop_replacement_dep = nn.Parameter(
            torch.randn(self.input_size + self.args["tag_emb_dim"])
            / np.sqrt(self.input_size + self.args["tag_emb_dim"])
        )
        self.drop_replacement_pos = nn.Parameter(torch.randn(self.input_size) / np.sqrt(self.input_size))

        self.upos_hid = nn.Linear(self.input_size, self.args["deep_biaff_hidden_dim"])
        self.upos_clf = nn.Linear(self.args["deep_biaff_hidden_dim"], len(vocab["upos"]))

        self.upos_emb_matrix_ner = nn.Parameter(
            torch.rand(len(vocab["upos"]), self.args["tag_emb_dim"]), requires_grad=True
        )
        self.upos_emb_matrix_dep = nn.Parameter(
            torch.rand(len(vocab["upos"]), self.args["tag_emb_dim"]), requires_grad=True
        )
        self.upos_clf.weight.data.zero_()
        self.upos_clf.bias.data.zero_()

        self.dep_hid = nn.Linear(
            self.input_size + self.args["tag_emb_dim"], self.input_size + self.args["tag_emb_dim"]
        )

        # classifiers
        self.unlabeled = DeepBiaffineScorer(
            self.input_size + self.args["tag_emb_dim"],
            self.input_size + self.args["tag_emb_dim"],
            self.args["deep_biaff_hidden_dim"],
            1,
            pairwise=True,
            dropout=args["dropout"],
        )
        self.deprel = DeepBiaffineScorer(
            self.input_size + self.args["tag_emb_dim"],
            self.input_size + self.args["tag_emb_dim"],
            self.args["deep_biaff_hidden_dim"],
            len(vocab["deprel"]),
            pairwise=True,
            dropout=args["dropout"],
        )
        if args["linearization"]:
            self.linearization = DeepBiaffineScorer(
                self.input_size + self.args["tag_emb_dim"],
                self.input_size + self.args["tag_emb_dim"],
                self.args["deep_biaff_hidden_dim"],
                1,
                pairwise=True,
                dropout=args["dropout"],
            )
        if args["distance"]:
            self.distance = DeepBiaffineScorer(
                self.input_size + self.args["tag_emb_dim"],
                self.input_size + self.args["tag_emb_dim"],
                self.args["deep_biaff_hidden_dim"],
                1,
                pairwise=True,
                dropout=args["dropout"],
            )

        self.ner_tag_clf = nn.Linear(self.input_size + self.args["tag_emb_dim"], len(self.vocab["ner_tag"]))
        self.ner_tag_clf.bias.data.zero_()
        # criterion
        self.crit_ner = CRFLoss(len(self.vocab["ner_tag"]))
        self.crit_dep = nn.CrossEntropyLoss(ignore_index=-1, reduction="sum")  # ignore padding
        self.crit_pos = nn.CrossEntropyLoss(ignore_index=1)

        # Dropout
        self.drop_ner = nn.Dropout(args["dropout"])
        self.worddrop_ner = WordDropout(args["word_dropout"])
        self.lockeddrop = LockedDropout(args["word_dropout"])

        self.drop_pos = nn.Dropout(args["dropout"])
        self.worddrop_pos = WordDropout(args["word_dropout"])

        self.drop_dep = nn.Dropout(args["dropout"])
        self.worddrop_dep = WordDropout(args["word_dropout"])

    def pos_forward(self, tokens_phobert, first_subword, sentlens, use_soft_pos=False, upos=None):
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs = []
        phobert_emb = self.phobert(tokens_phobert)[2][-1]
        phobert_emb = torch.cat(
            [torch.index_select(phobert_emb[i], 0, first_subword[i]).unsqueeze(0) for i in range(phobert_emb.size(0))],
            dim=0,
        )
        if torch.cuda.is_available():
            phobert_emb = phobert_emb.cuda()
        phobert_emb = pack(phobert_emb)
        inputs += [phobert_emb]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, phobert_emb.batch_sizes), batch_first=True)[0]

        inputs_pos = inputs[0].data
        inputs_pos = self.worddrop_pos(inputs_pos, self.drop_replacement_pos)

        upos_hid = F.relu(self.upos_hid(inputs_pos))
        upos_pred = self.upos_clf(self.drop_pos(upos_hid))

        preds = [pad(upos_pred).max(2)[1]]
        if use_soft_pos is False:
            upos = pack(upos).data
            loss = self.crit_pos(upos_pred.view(-1, upos_pred.size(-1)), upos.view(-1))
            return loss, preds
        else:
            return inputs, F.softmax(pad(upos_pred), dim=-1)

    def ner_forward(self, tokens_phobert, first_subword, word_mask, sentlens, tags):
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs, pos_dis = self.pos_forward(tokens_phobert, first_subword, sentlens, use_soft_pos=True)
        phobert_emb = inputs[0]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, phobert_emb.batch_sizes), batch_first=True)[0]

        upos_embed_matrix_dup = self.upos_emb_matrix_ner.repeat(pos_dis.size(0), 1, 1)
        pos_emb = torch.matmul(pos_dis, upos_embed_matrix_dup)
        pos_emb = pack(pos_emb)
        inputs += [pos_emb]

        inputs = torch.cat([x.data for x in inputs], 1)
        inputs = self.worddrop_ner(inputs, self.drop_replacement_ner)
        ner_pred = self.ner_tag_clf(inputs)

        logits = pad(F.relu(ner_pred)).contiguous()
        loss, trans = self.crit_ner(logits, word_mask, tags)
        return loss, logits

    def dep_forward(
        self, tokens_phobert, first_subword, word_mask, number_of_words, sentlens, head, deprel, eval=False
    ):
        def pack(x):
            return pack_padded_sequence(x, sentlens, batch_first=True)

        inputs, pos_dis = self.pos_forward(tokens_phobert, first_subword, sentlens, use_soft_pos=True)
        phobert_emb = inputs[0]

        def pad(x):
            return pad_packed_sequence(PackedSequence(x, phobert_emb.batch_sizes), batch_first=True)[0]

        upos_embed_matrix_dup = self.upos_emb_matrix_dep.repeat(pos_dis.size(0), 1, 1)
        pos_emb = torch.matmul(pos_dis, upos_embed_matrix_dup)
        pos_emb = pack(pos_emb)
        inputs += [pos_emb]

        inputs = torch.cat([x.data for x in inputs], 1)
        inputs = self.worddrop_dep(inputs, self.drop_replacement_dep)

        hidden_out = self.dep_hid(inputs)
        hidden_out = pad(hidden_out)

        unlabeled_scores = self.unlabeled(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
        deprel_scores = self.deprel(self.drop_dep(hidden_out), self.drop_dep(hidden_out))

        if self.args["linearization"] or self.args["distance"]:
            head_offset = torch.arange(first_subword.size(1), device=unlabeled_scores.device).view(1, 1, -1).expand(
                first_subword.size(0), -1, -1
            ) - torch.arange(first_subword.size(1), device=unlabeled_scores.device).view(1, -1, 1).expand(
                first_subword.size(0), -1, -1
            )

        if self.args["linearization"]:
            lin_scores = self.linearization(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
            unlabeled_scores += F.logsigmoid(lin_scores * torch.sign(head_offset).float()).detach()

        if self.args["distance"]:
            dist_scores = self.distance(self.drop_dep(hidden_out), self.drop_dep(hidden_out)).squeeze(3)
            dist_pred = 1 + F.softplus(dist_scores)
            dist_target = torch.abs(head_offset)
            dist_kld = -torch.log((dist_target.float() - dist_pred) ** 2 / 2 + 1)
            unlabeled_scores += dist_kld.detach()

        diag = torch.eye(unlabeled_scores.size(-1), dtype=torch.bool, device=unlabeled_scores.device).unsqueeze(0)
        unlabeled_scores.masked_fill_(diag, -float("inf"))

        preds = []
        if eval is False:
            unlabeled_scores = unlabeled_scores[:, 1:, :]  # exclude attachment for the root symbol
            unlabeled_scores = unlabeled_scores.masked_fill(word_mask.unsqueeze(1), -float("inf"))
            unlabeled_target = head.masked_fill(word_mask[:, 1:], -1)
            loss = self.crit_dep(
                unlabeled_scores.contiguous().view(-1, unlabeled_scores.size(2)), unlabeled_target.view(-1)
            )

            deprel_scores = deprel_scores[:, 1:]  # exclude attachment for the root symbol
            deprel_scores = torch.gather(
                deprel_scores, 2, head.unsqueeze(2).unsqueeze(3).expand(-1, -1, -1, len(self.vocab["deprel"]))
            ).view(-1, len(self.vocab["deprel"]))
            deprel_target = deprel.masked_fill(word_mask[:, 1:], -1)
            loss += self.crit_dep(deprel_scores.contiguous(), deprel_target.view(-1))

            if self.args["linearization"]:
                # lin_scores = lin_scores[:, 1:].masked_select(goldmask)
                lin_scores = torch.gather(lin_scores[:, 1:], 2, head.unsqueeze(2)).view(-1)
                lin_scores = torch.cat([-lin_scores.unsqueeze(1) / 2, lin_scores.unsqueeze(1) / 2], 1)
                lin_target = torch.gather((head_offset[:, 1:] > 0).long(), 2, head.unsqueeze(2))
                loss += self.crit_dep(lin_scores.contiguous(), lin_target.view(-1))

            if self.args["distance"]:
                dist_kld = torch.gather(dist_kld[:, 1:], 2, head.unsqueeze(2))
                loss -= dist_kld.sum()

            loss /= number_of_words  # number of words
        else:
            loss = 0
            preds.append(F.log_softmax(unlabeled_scores, 2).detach().cpu().numpy())
            preds.append(deprel_scores.max(3)[1].detach().cpu().numpy())
        return loss, preds

    def forward(
        self,
        dep_tokens_phobert,
        dep_first_subword,
        dep_word_mask,
        dep_number_of_words,
        dep_head,
        dep_deprel,
        dep_sentlens,
        pos_tokens_phobert,
        pos_first_subword,
        pos_upos,
        pos_sentlens,
        ner_tokens_phobert,
        ner_first_subword,
        ner_word_mask,
        ner_sentlens,
        ner_tag,
        lambda_pos=1.0,
        lambda_ner=1.0,
        lambda_dep=1.0,
    ):
        loss_pos, preds_pos = self.pos_forward(pos_tokens_phobert, pos_first_subword, pos_sentlens, False, pos_upos)
        loss_ner, preds_ner = self.ner_forward(
            ner_tokens_phobert, ner_first_subword, ner_word_mask, ner_sentlens, ner_tag
        )
        loss_dep, preds_dep = self.dep_forward(
            dep_tokens_phobert,
            dep_first_subword,
            dep_word_mask,
            dep_number_of_words,
            dep_sentlens,
            dep_head,
            dep_deprel,
        )

        loss = lambda_pos * loss_pos + lambda_ner * loss_ner + lambda_dep * loss_dep
        preds = preds_pos + [preds_ner] + preds_dep
        return loss_pos, loss_ner, loss_dep, loss, preds



import torch


class Trainer:
    def change_lr(self, new_lr):
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = new_lr

    def save(self, filename):
        savedict = {"model": self.model.state_dict(), "optimizer": self.optimizer.state_dict()}
        torch.save(savedict, filename)

    def load(self, filename):
        savedict = torch.load(filename, lambda storage, loc: storage)

        self.model.load_state_dict(savedict["model"])
        if self.args["mode"] == "train":
            self.optimizer.load_state_dict(savedict["optimizer"])
